<?php

namespace Drupal\content_moderation_reviewer\Plugin\Field\FieldWidget;

use Drupal\Component\Utility\NestedArray;
use Drupal\content_moderation\ContentModerationState;
use Drupal\content_moderation_reviewer\Form\EntityModerationReviewerForm;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\Element\EntityAutocomplete;
use Drupal\Core\Field\EntityReferenceFieldItemListInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\WidgetBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\user\UserInterface;
use Symfony\Component\Validator\ConstraintViolationInterface;

/**
 * Plugin implementation of the 'entity_reference_autocomplete' widget.
 *
 * @FieldWidget(
 *   id = "content_moderation_reviewer_autocomplete",
 *   label = @Translation("Reviewer Autocomplete"),
 *   description = @Translation("An autocomplete text field."),
 *   field_types = {
 *     "entity_reference"
 *   }
 * )
 */
class ReviewerAutocomplete extends WidgetBase {

  protected static function ajaxWrapperId($field_name, array $element) {
    return $field_name . '-content-moderation-reviewer';
  }

  /**
   * @return \Drupal\content_moderation\ModerationInformation
   */
  protected function getModerationInformation() {
    return \Drupal::service('content_moderation.moderation_information');
  }

  public static function getToStateId($default_state, FormStateInterface $form_state) {
    if (!empty($form_state->getUserInput()['moderation_state'][0]['state'])) {
      $to_state = $form_state->getUserInput()['moderation_state'][0]['state'];
      return ['user_input', $to_state];
    }
    elseif ($form_state->hasValue(['moderation_state', 0, 'state'])) {
      $to_state = $form_state->getValue(['moderation_state', 0, 'state']);
      return ['form_values', $to_state];
    }
    else {
      $to_state = $default_state;
      return ['default_state', $to_state];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {
    assert($items instanceof EntityReferenceFieldItemListInterface);
    $build = ['target_id' => []];
    $entity = $form_state->getFormObject()->getEntity();
    assert($entity instanceof ContentEntityInterface);

    if ($workflow = $this->getModerationInformation()->getWorkflowForEntity($entity)) {
      $from_state = $entity->get('moderation_state')->value;
      list ($state_source, $to_state) = $this->getToStateId($from_state, $form_state);

      $form['moderation_state']['#attributes']['id'] = static::ajaxWrapperId('content_moderation_reviewer', $form);

      $build['target_id']['#type'] = 'value';
      $build['target_id']['#element_validate'] = [[__CLASS__, 'elementValidate']];
      $form['#process'][] = [__CLASS__, 'process'];

      // Don't allow to set a reviewable once content is published.
      $to_state_object = $workflow->getTypePlugin()->getState($to_state);
      assert($to_state_object instanceof ContentModerationState);
      if ($to_state_object->isPublishedState()) {
        return $build;
      }

      $build['target_id']['#type'] = 'textfield';
      $build['target_id']['#title'] = $items->getFieldDefinition()->getLabel();

      $referenced_entities = $items->referencedEntities();
      $build['target_id']['#default_value'] = isset($referenced_entities[$delta]) ? static::userToAutocompleteLabel($referenced_entities[$delta]) : '';

      $build['target_id']['#autocomplete_route_name'] = 'content_moderation_reviewer.autocomplete';
      $build['target_id']['#autocomplete_route_parameters'] = [
        'workflow_id' => $workflow->id(),
        'from_state' => $from_state,
        'to_state' => $to_state,
      ];
    }

    return $build;
  }

  /**
   * Resets the content_moderation_reviewer on ajax requests.
   */
  public static function elementValidate(&$element, FormStateInterface $form_state, &$form) {
    // Don't manipulate the form data when there is a real submission going on.
    if ($form_state->isSubmitted()) {
      return;
    }

    $entity = $form_state->getFormObject()->getEntity();
    assert($entity instanceof ContentEntityInterface);
    $from_state_id = $entity->get('moderation_state')->value;
    $to_state_id = $form_state->getValue(['moderation_state', 0, 'value']);
    if ($from_state_id !== $to_state_id) {
      NestedArray::setValue($form_state->getUserInput(), $element['#parents'], '');
      NestedArray::setValue($form_state->getValues(), $element['#parents'], '');
    }
  }

  public static function userToAutocompleteLabel(UserInterface $user) {
    return "{$user->label()} ({$user->id()})";
  }

  public static function process($element) {
    $element['moderation_state']['content_moderation_reviewer'] = $element['content_moderation_reviewer'];

    $element['moderation_state']['widget'][0]['state']['#ajax'] = [
      'callback' => [__CLASS__, 'updateWidget'],
      'wrapper' => static::ajaxWrapperId('content_moderation_reviewer', $element),
    ];

    unset($element['content_moderation_reviewer']);

    return $element;
  }

  public static function updateWidget($element, FormStateInterface $form_state) {
    if ($form_state->getFormObject() instanceof EntityModerationReviewerForm) {
      return $element['moderation_state']['content_moderation_reviewer'];
    }
    else {
      return $element['moderation_state'];
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function isApplicable(FieldDefinitionInterface $field_definition) {
    return parent::isApplicable($field_definition) && $field_definition->getFieldStorageDefinition()->getName() === 'content_moderation_reviewer';
  }

    /**
   * {@inheritdoc}
   */
  public function errorElement(array $element, ConstraintViolationInterface $error, array $form, FormStateInterface $form_state) {
    return $element['target_id'];
  }

  /**
   * {@inheritdoc}
   */
  public function massageFormValues(array $values, array $form, FormStateInterface $form_state) {
    foreach ($values as $key => $value) {
      $values[$key]['target_id'] = EntityAutocomplete::extractEntityIdFromAutocompleteInput($value['target_id']);
    }

    return $values;
  }

}
